/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Mit Mr 28 13:21:21 CEST 2001
    copyright            : (C) 2001 by Arnold Krille
    email                : arnold@arnoldarts.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <stdlib.h>
#include <string>
#include <fstream>
#include <vector>
#include <stdio.h>
#include <getopt.h>

struct replace{
	std::string voraus;
	std::string ersetzt;
};

std::vector <replace> rules;
std::string syntax( std::ifstream *src, char *c );
void initReplace();
bool satzzeichen( char c );
bool satzzeichen( std::ifstream *src );
void addLine( std::string &add, std::string &to, bool wrapsafe = false );

int main(int argc, char *argv[])
{
	initReplace();
bool blankline=true,
     comment=false,
     comment1=false,
     str=false,
     strcom=false,
     strrm=false,
     charc=false,
     full=false,
     help=false,
     geometry=false,
     ps=false,
     zeichen=false, // ob das vorherige Zeichen ein Satzzeichen war.
     highlight=false,
     linenumbering=false,
     wrapsafe=false,
     inc=false // if we are inside an include-statement.
     ;
int  skip=17,
     tabspace=2,
     line=1,
     optcount=1,
     optch=0;
std::string tab="";

	for(int i=0; i<argc; i++)
		if( !strcmp(argv[i],"--help") ) { help=true; }



if(!help) // If Help isn't wanted:
{
	while( ( optch=getopt( argc,argv, "s:cCft:gphHnw" ) ) > 0 )
	switch( optch )
	{
		case 's': skip=atoi( optarg ); optcount+=2; break; // Skip n lines at the beginning of the file.
		case 'c': strcom=true; optcount++; break; // Print strings in comments in roman.
		case 'C': strrm=true; optcount++; break; // Print strings in roman.
		case 'f': full=true; optcount++; break; // Create a full LaTeX document.
		case 't': tabspace=atoi( optarg ); optcount+=2; break; // Set the tabspace width to n.
		case 'g': geometry=true; optcount++; break; // Use package geometry.
		case 'p': ps=true; optcount++; break; // Use Postscriptfonts.
		case 'h': highlight=true; optcount++; break; // Turn on syntax higlighting.
		case 'n': linenumbering=true; optcount++; break; // Print the linenumbers.
		case 'w': wrapsafe=true; optcount++; break; // Protect against linewrapping
	}

//	std::cerr<<"This program has been modified for Gentoo Linux"<<std::endl;
	std::cerr<<std::endl<<"Processing file "<<argv[optcount]<<std::endl;
std::ifstream file;
	file.open( argv[optcount] );

	// Skip lines
	std::cerr<<"  Skipping "<<skip<<" lines."<<std::endl;
	for(int i=0; i<skip; i++)
	{
	std::string temp;
		getline(file, temp);
	}

	// Set tabspace
	std::cerr<<"  Tabspace set to "<<tabspace<<" spaces."<<std::endl;
	for(int i=0; i<tabspace; i++)
		tab+="\\ ";

	// Enable-Messages
	if( highlight ) std::cerr<<"  Syntax-highlighting enabled."<<std::endl;
	if( highlight && !ps ) std::cerr<<"  * You should use Postscriptfonts since the normal TeX  *"<<std::endl
                                   <<"  * Computer Modern has no bold face for the fixedwidth. *"<<std::endl;
	if(full) std::cerr<<"  Full-LaTeX-output enabled."<<std::endl;
	if(ps) std::cerr<<"  PostScriptFonts enabled."<<std::endl;
	if(geometry) std::cerr<<"  Geometry-package enabled."<<std::endl;
	if(strcom) std::cerr<<"  String-constants in comments are typesetted in Roman too."<<std::endl;
	if(linenumbering) std::cerr<<"  Linenumbering is enabled."<<std::endl;
	if(wrapsafe) std::cerr<<"  Linewrap protection enabled."<<std::endl;

	// Ausgabestring initialisieren
std::string output="{\\ttfamily \\raggedright \\small\n";
std::string out;
	if(linenumbering) out+="001 ";

	// Here we go:
	while( !file.eof() )
	{
//	zeichen = satzzeichen( &file );
	char c;
		file.get( c );
		if( c!='\n' ) blankline=false; else { line++; inc=false; }

		if(highlight && !comment1 && !comment && !str && !charc && zeichen) out+=syntax( &file, &c );

		switch(c)
		{
			// * * * Single Characters * * *
			case '~': out+="\\textasciitilde "; break;
			case '^': out+="\\textasciicircum "; break;
			case '_': out+="\\underline\\ ";  break;
			case '$':
			case '%':
			case '&':
			case '{':
			case '}': out+='\\'; out+=c; break;
			case '<':
			case '>': out+=c; out+="{}"; break;
			case ' ': out+="\\ "; break;
			case 9: out+=tab; break;
			case '#':
				out+='\\'; out+=c;
				char tmp[ 8 ]; file.get( tmp, 8 );
				if( !strcmp( tmp, "include" ) ) inc=true;
				out+=tmp;
				break;

			// * * * - * * *
			case '-': out += "-{}"; break;

			// * * * Asterix * * *
			case '*': out+="$\\ast$";
				if( comment && !comment1 )
				{
					file.get(c);
					if(c=='/') { out+="/}"; comment=false; }
						else { if(c=='\n') out+="\\\\"; }
					if( !( (c=='*')||(c=='/') )) out+=c;
				}
				break;

			// * * * Gnsefchen * * *
			case '"':
				if( ((!comment && !comment1) || strcom) && !charc && !inc && strrm )
				{
					if( !str ) { out+="\"{}\\textrm{"; str=true;}
						else { out+="}\"{}"; str=false; }
				} else out+="\"{}";
				break;

			// * * * Slash * * *
			case '/': file.get(c);
				if( c=='*' ) { out+="\\textsl{/$\\ast$"; comment=true; } else {
					if( c=='/' ) {
						if( !(comment1 || comment) && !str )
						{
							out+="\\textsl{//"; comment1=true;
						}
							else out+="//"; }
					else { out+='/'; file.unget(); }
				}
				break;

			// * * * Linebreak * * *
			case '\n':
				if(blankline) out+="\\ ";
				if(comment1) { out+="}"; comment1=false; }
				addLine( out, output, wrapsafe );
				output+="\\\\"; output+=c; blankline=true;
				char str[10];
				sprintf(str,"%03d ",line);
				if(linenumbering) out+=str;
				break;

			// * * * Backslash * * *
			case '\\': out+="$\\backslash$"; file.get(c);
				if(c=='\\') out+="$\\backslash$";
					else {
						if(c=='"') out+="''";
							else {
								if(c=='\'') out+=c;
									else file.unget();
								}
						}
				break;

			// * * * Characterconstants * * *
			case '\'': out+=c;
				//if( !str ) {
					if( !charc && !comment && !comment1 )
					{ /*out+="\\textbf{charc=true}";*/ charc=true; }
					else { /*out+="\\textbf{charc=false}";*/ charc=false;}
				//}
				break;
			default: out+=c;
				break;
			// It's a test - Thanks to Kuang-che Wu for the fix!
		}
		zeichen = satzzeichen( c );
	}
	// if the file doesn't end with a newline we need to add the last line
	addLine( out, output, wrapsafe );
	output+=" }\n\\normalfont\\normalsize\n\n";

	// * * * Full LaTeX-Files:
	if(full)
	{
	std::string tmp="";
		tmp+="\\documentclass[a4paper,10pt]{article}\n";
		tmp+="\\usepackage[T1]{fontenc}\n";
		tmp+="\\usepackage{isolatin1,german}\n";
		if(geometry)
			tmp+="\\usepackage[left=2cm,right=2cm,top=2cm,bottom=2cm]{geometry}";
		if(ps)
			tmp+="\\usepackage{times,helvet,courier}\n";
		tmp+="\\parindent0cm\n";
		tmp+="\\parskip1.5ex\n";
		tmp+="\\begin{document}\n";
		tmp+=output;
		tmp+="\\end{document}\n";
		output=tmp;
	}

	// * * * Output * * *
	std::cout<<output;
	file.close();
	std::cerr<<"Everything went well."<<std::endl<<std::endl;
}
else // * * * Helptext * * *
{
	std::cerr<<std::endl<<"This is C++2LaTeX (version "<<VERSION<<")"<<std::endl<<std::endl
			<<"Usage: "<<argv[0]<<" [options] <filename>"<<std::endl<<std::endl
			<<"<filename> is the name of the file you want to convert."<<std::endl
			<<"The resulting LaTeX-code comes to stdout which you can redirect to a file or pipe to LaTeX."<<std::endl<<std::endl
			<<"Possible Options are:"<<std::endl
			<<"    -c          : Controls wether String constants in Comments are typesetted in roman or not."<<std::endl
			<<"                  Default is not."<<std::endl
			<<"    -C          : Prints string constants in roman."<<std::endl
			<<"    -f          : The output is not only a text to insert into an existing LaTeX-document,"<<std::endl
			<<"                  but a full document with header and so on."<<std::endl
			<<"                  Default is not."<<std::endl
			<<"       -g       : If a full LaTeX-file is created: Adds geometry-package with 2cm margin on all sides"<<std::endl
			<<"                  Default is off."<<std::endl
			<<"       -p       : If a full LaTeX-file is created: Adds PostScript fonts {times,helvet,courier}."<<std::endl
			<<"                  Default is off. courier is used for the code, times for the string constants."<<std::endl
			<<"    -h          : Enables Syntax Highlighting."<<std::endl
			<<"                  Now implemented for C++."<<std::endl
			<<"    -s <number> : Controls how many lines are to be skipped at the beginning of the file."<<std::endl
			<<"                  Default are 17 lines which is the normal KDevelop-header."<<std::endl
			<<"    -t <number> : Controls how many spaces should be inserted for a tabulator."<<std::endl
			<<"                  Default is 2"<<std::endl
			<<"    -n          : Turn on Line numbering."<<std::endl
			<<"    -w          : Make sure long lines don't wrap."<<std::endl
			<<"    --help      : Shows this helpmessage."<<std::endl<<std::endl;
}

	std::cerr<<"(c)2001,2002,2003 by Arnold Krille <arnold@arnoldarts.de>"<<std::endl
	         <<"              > http://www.arnoldarts.de/ <"<<std::endl
	         <<"Helpers: Richard Garand (cpp2latex for gentoo, linewrapping)"<<std::endl
	         <<"         Eduard Brcker  (linenumbering)"<<std::endl;

  return EXIT_SUCCESS;
}

/** Syntaxhighlighting Funktion */
std::string syntax( std::ifstream *src, char *c )
{
std::string out="";
bool stimmt=false;
	for(int i=0; i<int(rules.size()); i++)
	{
int j=0;
		if( *c==rules[i].voraus[0] )
		{
			do
			{
				stimmt=false;
				j++;
				src->get( *c );
				if( *c==rules[i].voraus[j] ) stimmt=true;
				if( (j==int(rules[i].voraus.size())-1) && stimmt && satzzeichen( src ) ) { src->get( *c ); return rules[i].ersetzt; }
			}
			while (stimmt);
		}
		for(int k=0; k<=j; k++)
		{
			src->unget();
		}
		src->get( *c );
	}
	return out;
}

/** initialize the replacementrules  */
void initReplace()
{
replace dummy;
	dummy.voraus = "and"; dummy.ersetzt = "\\textbf{and}"; rules.push_back(dummy);
	dummy.voraus = "and_eq"; dummy.ersetzt = "\\textbf{and\\underline\\ eq}"; rules.push_back(dummy);
	dummy.voraus = "asm"; dummy.ersetzt = "\\textbf{asm}"; rules.push_back(dummy);
	dummy.voraus = "auto"; dummy.ersetzt = "\\textbf{auto}"; rules.push_back(dummy);
	dummy.voraus = "bitand"; dummy.ersetzt = "\\textbf{bitand}"; rules.push_back(dummy);
	dummy.voraus = "bitor"; dummy.ersetzt = "\\textbf{bitor}"; rules.push_back(dummy);
	dummy.voraus = "bool"; dummy.ersetzt = "\\textbf{bool}"; rules.push_back(dummy);
	dummy.voraus = "break"; dummy.ersetzt = "\\textbf{break}"; rules.push_back(dummy);
	dummy.voraus = "case"; dummy.ersetzt = "\\textbf{case}"; rules.push_back(dummy);
	dummy.voraus = "catch"; dummy.ersetzt = "\\textbf{catch}"; rules.push_back(dummy);
	dummy.voraus = "char"; dummy.ersetzt = "\\textbf{char}"; rules.push_back(dummy);
	dummy.voraus = "class"; dummy.ersetzt = "\\textbf{class}"; rules.push_back(dummy);
	dummy.voraus = "comp"; dummy.ersetzt = "\\textbf{comp}"; rules.push_back(dummy);
	dummy.voraus = "const"; dummy.ersetzt = "\\textbf{const}"; rules.push_back(dummy);
	dummy.voraus = "const_case"; dummy.ersetzt = "\\textbf{const\\underline\\ case}"; rules.push_back(dummy);
	dummy.voraus = "continue"; dummy.ersetzt = "\\textbf{continue}"; rules.push_back(dummy);
	dummy.voraus = "default"; dummy.ersetzt = "\\textbf{default}"; rules.push_back(dummy);
	dummy.voraus = "delete"; dummy.ersetzt = "\\textbf{delete}"; rules.push_back(dummy);
	dummy.voraus = "do"; dummy.ersetzt = "\\textbf{do}"; rules.push_back(dummy);
	dummy.voraus = "double"; dummy.ersetzt = "\\textbf{double}"; rules.push_back(dummy);
	dummy.voraus = "dynamic_cast"; dummy.ersetzt = "\\textbf{dynamic\\underline\\ cast}"; rules.push_back(dummy);
	dummy.voraus = "else"; dummy.ersetzt = "\\textbf{else}"; rules.push_back(dummy);
	dummy.voraus = "enum"; dummy.ersetzt = "\\textbf{enum}"; rules.push_back(dummy);
	dummy.voraus = "explicit"; dummy.ersetzt = "\\textbf{explicit}"; rules.push_back(dummy);
	dummy.voraus = "extern"; dummy.ersetzt = "\\textbf{extern}"; rules.push_back(dummy);
	dummy.voraus = "false"; dummy.ersetzt = "\\textbf{false}"; rules.push_back(dummy);
	dummy.voraus = "float"; dummy.ersetzt = "\\textbf{float}"; rules.push_back(dummy);
	dummy.voraus = "for"; dummy.ersetzt = "\\textbf{for}"; rules.push_back(dummy);
	dummy.voraus = "friend"; dummy.ersetzt = "\\textbf{friend}"; rules.push_back(dummy);
	dummy.voraus = "goto"; dummy.ersetzt = "\\textbf{goto}"; rules.push_back(dummy);
	dummy.voraus = "if"; dummy.ersetzt = "\\textbf{if}"; rules.push_back(dummy);
	dummy.voraus = "inline"; dummy.ersetzt = "\\textbf{inline}"; rules.push_back(dummy);
	dummy.voraus = "int"; dummy.ersetzt = "\\textbf{int}"; rules.push_back(dummy);
	dummy.voraus = "long"; dummy.ersetzt = "\\textbf{long}"; rules.push_back(dummy);
	dummy.voraus = "mutable"; dummy.ersetzt = "\\textbf{mutable}"; rules.push_back(dummy);
	dummy.voraus = "namespace"; dummy.ersetzt = "\\textbf{namespace}"; rules.push_back(dummy);
	dummy.voraus = "new"; dummy.ersetzt = "\\textbf{new}"; rules.push_back(dummy);
	dummy.voraus = "not"; dummy.ersetzt = "\\textbf{not}"; rules.push_back(dummy);
	dummy.voraus = "not_eq"; dummy.ersetzt = "\\textbf{not\\underline\\ eq}"; rules.push_back(dummy);
	dummy.voraus = "operator"; dummy.ersetzt = "\\textbf{operator}"; rules.push_back(dummy);
	dummy.voraus = "or"; dummy.ersetzt = "\\textbf{or}"; rules.push_back(dummy);
	dummy.voraus = "or_eq"; dummy.ersetzt = "\\textbf{or\\underline\\ eq}"; rules.push_back(dummy);
	dummy.voraus = "private"; dummy.ersetzt = "\\textbf{private}"; rules.push_back(dummy);
	dummy.voraus = "protected"; dummy.ersetzt = "\\textbf{protected}"; rules.push_back(dummy);
	dummy.voraus = "public"; dummy.ersetzt = "\\textbf{public}"; rules.push_back(dummy);
	dummy.voraus = "register"; dummy.ersetzt = "\\textbf{register}"; rules.push_back(dummy);
	dummy.voraus = "reinterpret_cast"; dummy.ersetzt = "\\textbf{reinterpret\\underline\\ cast}"; rules.push_back(dummy);
	dummy.voraus = "return"; dummy.ersetzt = "\\textbf{return}"; rules.push_back(dummy);
	dummy.voraus = "short"; dummy.ersetzt = "\\textbf{short}"; rules.push_back(dummy);
	dummy.voraus = "signed"; dummy.ersetzt = "\\textbf{signed}"; rules.push_back(dummy);
	dummy.voraus = "sizeof"; dummy.ersetzt = "\\textbf{sizeof}"; rules.push_back(dummy);
	dummy.voraus = "static"; dummy.ersetzt = "\\textbf{static}"; rules.push_back(dummy);
	dummy.voraus = "static_cast"; dummy.ersetzt = "\\textbf{static\\underline\\ cast}"; rules.push_back(dummy);
	dummy.voraus = "struct"; dummy.ersetzt = "\\textbf{struct}"; rules.push_back(dummy);
	dummy.voraus = "switch"; dummy.ersetzt = "\\textbf{switch}"; rules.push_back(dummy);
	dummy.voraus = "template"; dummy.ersetzt = "\\textbf{template}"; rules.push_back(dummy);
	dummy.voraus = "this"; dummy.ersetzt = "\\textbf{this}"; rules.push_back(dummy);
	dummy.voraus = "throw"; dummy.ersetzt = "\\textbf{throw}"; rules.push_back(dummy);
	dummy.voraus = "true"; dummy.ersetzt = "\\textbf{true}"; rules.push_back(dummy);
	dummy.voraus = "try"; dummy.ersetzt = "\\textbf{try}"; rules.push_back(dummy);
	dummy.voraus = "typedef"; dummy.ersetzt = "\\textbf{typedef}"; rules.push_back(dummy);
	dummy.voraus = "typeid"; dummy.ersetzt = "\\textbf{typeid}"; rules.push_back(dummy);
	dummy.voraus = "typename"; dummy.ersetzt = "\\textbf{typename}"; rules.push_back(dummy);
	dummy.voraus = "union"; dummy.ersetzt = "\\textbf{union}"; rules.push_back(dummy);
	dummy.voraus = "unsigned"; dummy.ersetzt = "\\textbf{unsigned}"; rules.push_back(dummy);
	dummy.voraus = "virtual"; dummy.ersetzt = "\\textbf{virtual}"; rules.push_back(dummy);
	dummy.voraus = "void"; dummy.ersetzt = "\\textbf{void}"; rules.push_back(dummy);
	dummy.voraus = "volatile"; dummy.ersetzt = "\\textbf{volatile}"; rules.push_back(dummy);
	dummy.voraus = "wchar_t"; dummy.ersetzt = "\\textbf{wchar\\underline\\ t}"; rules.push_back(dummy);
	dummy.voraus = "while"; dummy.ersetzt = "\\textbf{while}"; rules.push_back(dummy);
	dummy.voraus = "xor"; dummy.ersetzt = "\\textbf{xor}"; rules.push_back(dummy);
	dummy.voraus = "xor_eq"; dummy.ersetzt = "\\textbf{xor\\underline\\ eq}"; rules.push_back(dummy);

}

/** Gibt true zurck, wenn es sich beim bergebenen Zeichen um ein Satzzeichen handelt. */
bool satzzeichen( char c ){
	// zugelassene Sonderzeichen:
	switch(c)
	{
		case '\'':
			return false;
		break;
	}
	// Zahlen:
	if( c >= 48 && c <= 57) return false;
	// kleine Buchstaben:
	if( c >= 97 && c <= 122) return false;
	// Groe Buchstaben:
	if( c >= 65 && c <= 90) return false;
return true;
}

/** Gibt true zurck, wenn es sich beim nchsten Zeichen um ein Satzzeichen handelt. */
bool satzzeichen( std::ifstream *src ){
char c;
	src->get( c );
	src->unget();
	return satzzeichen( c );
}

/** add add to to, if wrapsafe is true use mbox on long lines, clear add */
void addLine( std::string &add, std::string &to, bool wrapsafe )
{
	if ( add.length() > 65 && wrapsafe ) {
		to += "\\mbox{";
		to += add;
		to += "}";
	} else
		to += add;
	add = "";
}
